//----------------------------------------------------------------------//
// NyanFi																//
//  J[sbJ[														//
//----------------------------------------------------------------------//
#include <vcl.h>
#pragma hdrstop
#include <tchar.h>
#include <algorithm>
#include <memory>
#include <Vcl.Clipbrd.hpp>
#include "UserFunc.h"
#include "UserMdl.h"
#include "Global.h"
#include "ColPicker.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TColorPicker *ColorPicker;

//---------------------------------------------------------------------------
__fastcall TColorPicker::TColorPicker(TComponent* Owner)
	: TForm(Owner)
{
}

//---------------------------------------------------------------------------
void __fastcall TColorPicker::FormCreate(TObject *Sender)
{
	ViewImage = NULL;

	set_ComboBoxText(FmtComboBox,
		_T("R,G,B\r\n")
		_T("H,S,V(B)\r\n")
		_T("H,S,L\r\n")
		_T("COLORREF\r\n")
		_T("Delphi`\r\n")
		_T("VB`\r\n")
		_T("HTML`\r\n")
		_T("RGB(x,x,x)\r\n")
		_T("RGB(d,d,d)\r\n")
		_T("萔V{\r\n")
		_T("BITMAP(32)\r\n")
		_T("BITMAP(128)\r\n")
		_T("BITMAP(256)\r\n"));
}
//---------------------------------------------------------------------------
void __fastcall TColorPicker::FormShow(TObject *Sender)
{
	IniFile->LoadPosInfo(this);

	FmtComboBox->ItemIndex	 = IniFile->ReadIntGen(_T("ColorPickerFmt"));
	Ave9pxCheckBox->Checked  = IniFile->ReadBoolGen(_T("ColorPickerAve9px"));
	CmpColCheckBox->Checked  = IniFile->ReadBoolGen(_T("ColorPickerCmpCol"));
	SafeColCheckBox->Checked = IniFile->ReadBoolGen(_T("ColorPickerSafeCol"));

	ColLabel->Font->Size = 9;
	ColImage->Picture->Bitmap->SetSize(ColImage->Width, ColImage->Width);
	TCanvas *cv = ColImage->Canvas;
	cv->Brush->Color = col_bgImage;
	cv->FillRect(ColImage->ClientRect);

	if (ViewImage) ViewImage->Cursor = UserModule->crSpuitTool;
}
//---------------------------------------------------------------------------
void __fastcall TColorPicker::FormClose(TObject *Sender, TCloseAction &Action)
{
	IniFile->SavePosInfo(this);

	IniFile->WriteIntGen(_T("ColorPickerFmt"),		FmtComboBox);
	IniFile->WriteBoolGen(_T("ColorPickerAve9px"),	Ave9pxCheckBox);
	IniFile->WriteBoolGen(_T("ColorPickerCmpCol"),	CmpColCheckBox);
	IniFile->WriteBoolGen(_T("ColorPickerSafeCol"),	SafeColCheckBox);

	if (ViewImage) ViewImage->Cursor = crDefault;
}

//---------------------------------------------------------------------------
//RGBHSL ϊ
//---------------------------------------------------------------------------
void __fastcall TColorPicker::RgbToHslI(TColor col, int *h, int *s, int *l)
{
	int cref = ColorToRGB(col);
	float r = GetRValue(cref)/255.0;
	float g = GetGValue(cref)/255.0;
	float b = GetBValue(cref)/255.0;

	float maxv = std::max(r, std::max(g, b));
	float minv = std::min(r, std::min(g, b));
	float h_f, s_f;
	float l_f = (maxv + minv)/2.0;

	if (maxv==minv) {
		h_f = s_f = 0.0;
	}
	else {
		float d = maxv - minv;
		if		(maxv==r) h_f = (g - b)/d;
		else if (maxv==g) h_f = (b - r)/d + 2.0;
		else			  h_f = (r - g)/d + 4.0;
		h_f *= 60.0;
		if (h_f<0.0) h_f += 360.0;

		s_f = ((l_f<0.5)? d/(maxv + minv) : d/(2.0 - maxv - minv)) * 100.0;
	}

	l_f *= 100.0;

	*h = (int)(h_f + 0.5);
	*s = (int)(s_f + 0.5);
	*l = (int)(l_f + 0.5);
}
//---------------------------------------------------------------------------
//RGBHSV ϊ
//---------------------------------------------------------------------------
void __fastcall TColorPicker::RgbToHsvI(TColor col, int *h, int *s, int *v)
{
	int cref = ColorToRGB(col);
	float r = GetRValue(cref)/255.0;
	float g = GetGValue(cref)/255.0;
	float b = GetBValue(cref)/255.0;

	float maxv = std::max(r, std::max(g, b));
	float minv = std::min(r, std::min(g, b));
	float h_f, s_f;
	float v_f = maxv * 100;

	if (maxv==minv) {
		h_f = s_f = 0.0;
	}
	else {
		float d = maxv - minv;
		if		(maxv==r) h_f = (g - b)/d;
		else if (maxv==g) h_f = (b - r)/d + 2.0;
		else			  h_f = (r - g)/d + 4.0;
		h_f *= 60.0;
		if (h_f<0.0) h_f += 360.0;

		s_f = d/maxv * 100.0;
	}

	*h = (int)(h_f + 0.5);
	*s = (int)(s_f + 0.5);
	*v = (int)(v_f + 0.5);
}

//---------------------------------------------------------------------------
//\̍XV
//---------------------------------------------------------------------------
void __fastcall TColorPicker::UpdateStt(int x, int y, float ratio)
{
	if (!ViewImage) return;

	Graphics::TBitmap *bmp = ViewImage->Picture->Bitmap;

	TRect rc = Rect(0, 0, bmp->Width, bmp->Height);
	if (ViewImage->Center)
		rc.Offset((ViewImage->ClientWidth - bmp->Width)/2, (ViewImage->ClientHeight - bmp->Height)/2);
	x -= rc.Left;
	y -= rc.Top;

	int c_size = ColImage->Width/3;
	TCanvas *cv = ColImage->Canvas;
	if (x<0 || x>=bmp->Width || y<0 || y>=bmp->Height) {
		ColLabel->Caption = "X___ Y___ (x___ y___)\nR___ G___ B___  #______\nH___ S___ V___ / H___ S___ L___";
		cv->Brush->Color  = col_bgImage;
		cv->FillRect(Rect(0, 0, c_size*3, c_size*3));
		Col2Panel->Color  = col_bgImage;
		ColEdit->Color	  = col_Invalid;
		ColEdit->Text	  = EmptyStr;
	}
	else {
		TColor col[9];
		for (int i=0; i<3; i++) {
			int xp	= i*c_size;
			int x_i = x + i - 1;
			for (int j=0; j<3; j++) {
				int yp	= j*c_size;
				int y_j = y + j - 1;
				cv->Brush->Color = (x_i>=0 && x_i<bmp->Width && y_j>=0 && y_j<bmp->Height)?
										bmp->Canvas->Pixels[x_i][y_j] : col_bgImage;
				col[i*3 + j] = cv->Brush->Color;
				TRect rc = Rect(xp, yp, xp + c_size, yp + c_size);
				cv->FillRect(rc);
			}
		}

		//Fl̕\
		TColor c = col[4];
		int cref = ColorToRGB(c);
		int r = GetRValue(cref);
		int g = GetGValue(cref);
		int b = GetBValue(cref);
		int h, s, v, l;
		UnicodeString msg;
		RgbToHsvI(c, &h, &s, &v);
		msg.sprintf(_T("X%d Y%d (x%d y%d)\nR%d G%d B%d  #%02x%02x%02x\nH%d S%d V%d"), 
						(int)(x/ratio + 0.5), (int)(y/ratio + 0.5),  x, y,  r, g, b,  r, g, b,  h, s, v);
		RgbToHslI(c, &h, &s, &l);
		msg.cat_sprintf(_T(" / H%d S%d L%d"), h, s, l);
		ColLabel->Caption = msg;
		//Ώۘg
		bool same_col = true;
		for (int i=1; i<9 && same_col; i++) same_col = (col[i]==col[i - 1]);
		cv->Brush->Color = RatioCol((h>=90 && h<270)? clRed : clAqua, (std::min(std::max(100 - l, 25), 75))/100.0);
		cv->FrameRect(same_col? Rect(0, 0, c_size*3, c_size*3) : Rect(c_size, c_size, c_size*2, c_size*2));

		//ԏ
		int r2 = r, g2 = g, b2 = b;
		//9_
		if (Ave9pxCheckBox->Checked) {
			int sum_r = 0, sum_g = 0, sum_b = 0;
			for (int i=0; i<9; i++) {
				int cref = ColorToRGB(col[i]);
				sum_r += GetRValue(cref);
				sum_g += GetGValue(cref);
				sum_b += GetBValue(cref);
			}
			r2 = sum_r/9;
			g2 = sum_g/9;
			b2 = sum_b/9;
		}
		//F
		if (CmpColCheckBox->Checked) {
			int x = std::max(r, std::max(g2, b2)) + std::min(r2, std::min(g2, b2));
			r2 = x - r2;
			g2 = x - g2;
			b2 = x - b2;
		}
		//Z[tJ[ߎ
		if (SafeColCheckBox->Checked) {
			r2 = (int)(r2/51.0 + 0.5) * 51;
			g2 = (int)(g2/51.0 + 0.5) * 51;
			b2 = (int)(b2/51.0 + 0.5) * 51;
		}

		//擾F
		TColor c2 = TColor(RGB(r2, g2, b2));
		int cref2 = ColorToRGB(c2);
		int h2, s2, v2;  RgbToHsvI(c2, &h2, &s2, &v2);
		int h3, s3, l2;  RgbToHslI(c2, &h3, &s3, &l2);
		Col2Panel->Color = c2;

		//Rs[
		UnicodeString colstr;
		bool is_bmp = false;

		switch (FmtComboBox->ItemIndex) {
		case 0: colstr.sprintf(_T("%d,%d,%d"),			r2, g2, b2);	break;
		case 1: colstr.sprintf(_T("%d,%d,%d"),			h2, s2, v2);	break;
		case 2: colstr.sprintf(_T("%d,%d,%d"),			h3, s3, l2);	break;
		case 3: colstr.sprintf(_T("0x%08x"),			cref2);			break;
		case 4: colstr.sprintf(_T("$00%02X%02X%02X"),	b2, g2, r2);	break;
		case 5: colstr.sprintf(_T("&H00%02X%02X%02X&"),	b2, g2, r2);	break;
		case 6: colstr.sprintf(_T("#%02x%02x%02x"),		r2, g2, b2);	break;
		case 7: colstr.sprintf(_T("RGB(0x%02x,0x%02x,0x%02x)"), r2, g2, b2);	break;
		case 8: colstr.sprintf(_T("RGB(%d,%d,%d)"),		r2, g2, b2);	break;
		case 9:
			colstr = ColorToString(c2);
			if (USTARTS_TS('0', colstr)) colstr = RGBToWebColorName(cref2);
			if (USTARTS_TS('#', colstr)) colstr.sprintf(_T("0x%08x"), cref2);
			break;
		default:	//10..12
			if (USTARTS_TS("BITMAP(", FmtComboBox->Text)) {
				colstr.sprintf(_T("R%d G%d B%d"), r2, g2, b2);
				is_bmp = true;
			}
		}

		ColEdit->Color = is_bmp? col_Invalid : scl_Window;
		ColEdit->Text  = colstr;
	}
}

//---------------------------------------------------------------------------
//ʂ̃Rs[
//---------------------------------------------------------------------------
void __fastcall TColorPicker::CopyColor()
{
	UnicodeString fmt = FmtComboBox->Text;
	//rbg}bv
	if (remove_top_s(fmt, _T("BITMAP("))) {
		int size = get_tkn(fmt, ')').ToIntDef(32);
		std::unique_ptr<Graphics::TBitmap> bmp(new Graphics::TBitmap());
		bmp->SetSize(size, size);
		TCanvas *cv = bmp->Canvas;
		cv->Brush->Color = Col2Panel->Color;
		cv->FillRect(Rect(0, 0, size, size));
		copy_to_Clipboard(bmp.get());
	}
	//
	else
		copy_to_Clipboard(ColEdit->Text);

	Close();
}

//---------------------------------------------------------------------------
void __fastcall TColorPicker::FormKeyDown(TObject *Sender, WORD &Key, TShiftState Shift)
{
	if (Key==VK_ESCAPE)
		Close(); 
	else
		SpecialKeyProc(this, Key, Shift, _T(HELPTOPIC_IV) _T("#ColorPicker"));
}
//---------------------------------------------------------------------------
